@@ -1,5 +1,5 @@
 /**********************************************************************
- * $Id: shp2pgsql.c 2513 2006-10-14 14:22:10Z mschaber $
+ * $Id: shp2pgsql.c 2717 2007-11-23 10:42:09Z mcayland $
  *
  * PostGIS - Spatial Types for PostgreSQL
  * http://postgis.refractions.net
@@ -138,7 +138,7 @@ static void print_wkb_int(int val);
 static void print_wkb_double(double val);
 
 static char rcsid[] =
-  "$Id: shp2pgsql.c 2513 2006-10-14 14:22:10Z mschaber $";
+  "$Id: shp2pgsql.c 2717 2007-11-23 10:42:09Z mcayland $";
 
 void *safe_malloc(size_t size)
 {
@@ -251,7 +251,8 @@ protect_quotes_string(char *str)
 	optr=result;
 	ptr=str;
 	while (*ptr) {
-		if ( *ptr == '\'' || *ptr == '\\' ) *optr++='\\';
+		if ( *ptr == '\\' ) *optr++='\\';
+                if ( *ptr == '\'') *optr++='\'';
 		*optr++=*ptr++;
 	}
 	*optr='\0';
@@ -421,16 +422,16 @@ main (int ARGC, char **ARGV)
 	 */
 	if(opt != 'a') CreateTable();
 
-	/*
-	 * Create GiST index if requested
-	 */
-	if(createindex) CreateIndex();
-
 	/*
 	 * Generate INSERT or COPY lines
 	 */
 	if(opt != 'p') LoadData();
 
+	/*
+	 * Create GiST index if requested
+	 */
+	if(createindex) CreateIndex();
+
 	printf("END;\n"); /* End the last transaction */
 
 
@@ -535,15 +536,15 @@ CreateTable(void)
 			{
 				printf ("int4");
 			}
-			else if  ( field_width <= 5 )
+			else if  ( field_width < 5 )
 			{
 				printf ("int2");
 			}
-			else if  ( field_width <= 10 )
+			else if  ( field_width < 10 )
 			{
 				printf ("int4");
 			}
-			else if  ( field_width <= 19 )
+			else if  ( field_width < 19 )
 			{
 				printf ("int8");
 			}
