@@ -1,5 +1,5 @@
 /**********************************************************************
- * $Id: lwgeom_functions_basic.c 2532 2006-12-01 22:16:44Z mleslie $
+ * $Id: lwgeom_functions_basic.c 2713 2007-10-29 13:44:29Z mcayland $
  *
  * PostGIS - Spatial Types for PostgreSQL
  * http://postgis.refractions.net
@@ -82,9 +82,9 @@ PG_FUNCTION_INFO_V1(LWGEOM_mem_size);
 Datum LWGEOM_mem_size(PG_FUNCTION_ARGS)
 {
 	PG_LWGEOM *geom = (PG_LWGEOM *) PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
-	size_t size = geom->size;
+	size_t size = VARSIZE(geom);
 	size_t computed_size = lwgeom_size(SERIALIZED_FORM(geom));
-	computed_size += 4; /* varlena size */
+	computed_size += VARHDRSZ; /* varlena size */
 	if ( size != computed_size )
 	{
 		elog(NOTICE, "varlena size (%lu) != computed size+4 (%lu)",
@@ -113,7 +113,7 @@ Datum LWGEOM_summary(PG_FUNCTION_ARGS)
 
 	/* create a text obj to return */
 	mytext = (text *) lwalloc(VARHDRSZ  + strlen(result) + 1);
-	VARATT_SIZEP(mytext) = VARHDRSZ + strlen(result) + 1;
+	SET_VARSIZE(mytext, VARHDRSZ + strlen(result) + 1);
 	VARDATA(mytext)[0] = '\n';
 	memcpy(VARDATA(mytext)+1, result, strlen(result) );
 
@@ -129,7 +129,7 @@ Datum postgis_version(PG_FUNCTION_ARGS)
 	char *ver = POSTGIS_VERSION;
 	text *result;
 	result = lwalloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -140,7 +140,7 @@ Datum postgis_lib_version(PG_FUNCTION_ARGS)
 	char *ver = POSTGIS_LIB_VERSION;
 	text *result;
 	result = lwalloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -151,7 +151,7 @@ Datum postgis_lib_build_date(PG_FUNCTION_ARGS)
 	char *ver = POSTGIS_BUILD_DATE;
 	text *result;
 	result = lwalloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -162,7 +162,7 @@ Datum postgis_scripts_released(PG_FUNCTION_ARGS)
 	char *ver = POSTGIS_SCRIPTS_VERSION;
 	text *result;
 	result = lwalloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -320,6 +320,8 @@ Datum LWGEOM_area_polygon(PG_FUNCTION_ARGS)
 	PG_LWGEOM *geom = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
 	LWGEOM_INSPECTED *inspected = lwgeom_inspect(SERIALIZED_FORM(geom));
 	LWPOLY *poly;
+        LWCURVEPOLY *curvepoly;
+        LWGEOM *tmp;
 	double area = 0.0;
 	int i;
 
@@ -329,9 +331,18 @@ Datum LWGEOM_area_polygon(PG_FUNCTION_ARGS)
 
 	for (i=0; i<inspected->ngeometries; i++)
 	{
-		poly = lwgeom_getpoly_inspected(inspected, i);
-		if ( poly == NULL ) continue;
-		area += lwgeom_polygon_area(poly);
+                tmp = lwgeom_getgeom_inspected(inspected, i);
+                if(lwgeom_getType(tmp->type) == POLYGONTYPE)
+                {
+		        poly = (LWPOLY *)tmp;
+		        area += lwgeom_polygon_area(poly);
+                }
+                else if(lwgeom_getType(tmp->type) == CURVEPOLYTYPE)
+                {
+                        curvepoly = (LWCURVEPOLY *)tmp;
+                        area += lwgeom_curvepolygon_area(curvepoly);
+                }
+                lwgeom_release(tmp);
 #ifdef PGIS_DEBUG
 		elog(NOTICE, " LWGEOM_area_polygon found a poly (%f)", area);
 #endif
@@ -1912,7 +1923,7 @@ Datum LWGEOM_accum(PG_FUNCTION_ARGS)
 			PG_RETURN_NULL();
 		}
 
-		result->size = nbytes;
+		SET_VARSIZE(result, nbytes);
 		result->ndim = 1;
 
 #if USE_VERSION > 72
@@ -1929,7 +1940,7 @@ Datum LWGEOM_accum(PG_FUNCTION_ARGS)
 #endif
 
 	} else {
-		oldsize = array->size;
+		oldsize = VARSIZE(array);
 		nbytes = oldsize + INTALIGN(geom->size);
 #ifdef PGIS_DEBUG
 		elog(NOTICE, "geom_accum: old array size: %d, adding %ld bytes (nelems=%d; nbytes=%u)", array->size, INTALIGN(geom->size), nelems, nbytes);
@@ -1950,7 +1961,7 @@ Datum LWGEOM_accum(PG_FUNCTION_ARGS)
 			ARR_DATA_PTR(result), (uchar *)ARR_DATA_PTR(result)-(uchar *)result);
 		elog(NOTICE, " next element @ %p", (uchar *)result+oldsize);
 #endif
-		result->size = nbytes;
+		SET_VARSIZE(result, nbytes);
 		memcpy(ARR_DIMS(result), &nelems, sizeof(int));
 #ifdef PGIS_DEBUG
 		elog(NOTICE, " writing next element starting @ %p",
@@ -2482,45 +2493,86 @@ PG_FUNCTION_INFO_V1(LWGEOM_envelope);
 Datum LWGEOM_envelope(PG_FUNCTION_ARGS)
 {
 	PG_LWGEOM *geom = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
-	BOX2DFLOAT4 box;
-	POINT2D *pts = lwalloc(sizeof(POINT2D)*5);
-	POINTARRAY *pa[1];
-	LWPOLY *poly;
+	BOX3D box;
 	int SRID;
+	POINTARRAY *pa;
 	PG_LWGEOM *result;
-	uchar *ser;
+	uchar *ser = NULL;
+
 
 	/* get bounding box  */
-	if ( ! getbox2d_p(SERIALIZED_FORM(geom), &box) )
+	if ( ! compute_serialized_box3d_p(SERIALIZED_FORM(geom), &box) )
 	{
 		/* must be the EMPTY geometry */
 		PG_RETURN_POINTER(geom);
 	}
-
+	
 	/* get geometry SRID */
 	SRID = lwgeom_getsrid(SERIALIZED_FORM(geom));
 
+	
+	/* 
+	 * Alter envelope type so that a valid geometry is always
+	 * returned depending upon the size of the geometry. The
+	 * code makes the following assumptions:
+	 *     - If the bounding box is a single point then return a
+	 *     POINT geometry
+	 *     - If the bounding box represents either a horizontal or
+	 *     vertical line, return a LINESTRING geometry
+	 *     - Otherwise return a POLYGON
+	 */
+
+
+	if (box.xmin == box.xmax &&
+	    box.ymin == box.ymax)
+        {
+                /* Construct and serialize point */
+                LWPOINT *point = make_lwpoint2d(SRID, box.xmin, box.ymin);
+                ser = lwpoint_serialize(point);
+        }
+	else if (box.xmin == box.xmax ||
+	         box.ymin == box.ymax)
+        {
+                LWLINE *line;
+                POINT2D *pts = palloc(sizeof(POINT2D)*2);
+
+                /* Assign coordinates to POINT2D array */
+                pts[0].x = box.xmin; pts[0].y = box.ymin;
+                pts[1].x = box.xmax; pts[1].y = box.ymax;
+
+                /* Construct point array */
+                pa = pointArray_construct((uchar *)pts, 0, 0, 2);	  
+
+                /* Construct and serialize linestring */
+                line = lwline_construct(SRID, NULL, pa);
+                ser = lwline_serialize(line);
+        }
+        else
+        {
+                LWPOLY *poly;
+	        POINT2D *pts = lwalloc(sizeof(POINT2D)*5);
+                BOX2DFLOAT4 box2d;
+                getbox2d_p(SERIALIZED_FORM(geom), &box2d);
+  		
+                /* Assign coordinates to POINT2D array */
+                pts[0].x = box2d.xmin; pts[0].y = box2d.ymin;
+                pts[1].x = box2d.xmin; pts[1].y = box2d.ymax;
+                pts[2].x = box2d.xmax; pts[2].y = box2d.ymax;
+                pts[3].x = box2d.xmax; pts[3].y = box2d.ymin;
+                pts[4].x = box2d.xmin; pts[4].y = box2d.ymin;
+
+                /* Construct point array */
+                pa = pointArray_construct((uchar *)pts, 0, 0, 5);
+
+                /* Construct polygon  */
+                poly = lwpoly_construct(SRID, box2d_clone(&box2d), 1, &pa);
+
+                /* Serialize polygon */
+                ser = lwpoly_serialize(poly);
+        }
+			
 	PG_FREE_IF_COPY(geom, 0);
 
-	/* Assign coordinates to POINT2D array */
-	pts[0].x = box.xmin; pts[0].y = box.ymin;
-	pts[1].x = box.xmin; pts[1].y = box.ymax;
-	pts[2].x = box.xmax; pts[2].y = box.ymax;
-	pts[3].x = box.xmax; pts[3].y = box.ymin;
-	pts[4].x = box.xmin; pts[4].y = box.ymin;
-
-	/* Construct point array */
-	pa[0] = lwalloc(sizeof(POINTARRAY));
-	pa[0]->serialized_pointlist = (uchar *)pts;
-	TYPE_SETZM(pa[0]->dims, 0, 0);
-	pa[0]->npoints = 5;
-
-	/* Construct polygon  */
-	poly = lwpoly_construct(SRID, box2d_clone(&box), 1, pa);
-
-	/* Serialize polygon */
-	ser = lwpoly_serialize(poly);
-
 	/* Construct PG_LWGEOM  */
 	result = PG_LWGEOM_construct(ser, SRID, 1);
 
@@ -2969,7 +3021,7 @@ Datum LWGEOM_asEWKT(PG_FUNCTION_ARGS)
 	PG_LWGEOM *lwgeom;
 	char *result_cstring;
 	int len;
-        char *result,*loc_wkt;
+    char *result,*loc_wkt;
 	/*char *semicolonLoc; */
 
 	init_pg_func();
@@ -2986,11 +3038,11 @@ Datum LWGEOM_asEWKT(PG_FUNCTION_ARGS)
 #endif
 	loc_wkt = result_cstring;
 
-	len = strlen(loc_wkt)+4;
-	result = palloc(len);
-	memcpy(result, &len, 4);
+	len = strlen(loc_wkt);
+	result = palloc(len + VARHDRSZ);
+    SET_VARSIZE(result, len + VARHDRSZ);
 
-	memcpy(result+4,loc_wkt, len-4);
+	memcpy(VARDATA(result), loc_wkt, len);
 
 	pfree(result_cstring);
 	PG_FREE_IF_COPY(lwgeom, 0);
