@@ -8,6 +8,7 @@
 #include "profile.h"
 #include "wktparse.h"
 #include "geos_c.h"
+#include "lwgeom_rtree.h"
 
 /*
  * Define this to have have many notices printed
@@ -45,6 +46,7 @@ Datum intersects(PG_FUNCTION_ARGS);
 Datum crosses(PG_FUNCTION_ARGS);
 Datum within(PG_FUNCTION_ARGS);
 Datum contains(PG_FUNCTION_ARGS);
+Datum covers(PG_FUNCTION_ARGS);
 Datum overlaps(PG_FUNCTION_ARGS);
 Datum isvalid(PG_FUNCTION_ARGS);
 Datum buffer(PG_FUNCTION_ARGS);
@@ -81,7 +83,7 @@ Datum postgis_geos_version(PG_FUNCTION_ARGS)
 	const char *ver = GEOSversion();
 	text *result;
 	result = (text *) palloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -142,7 +144,7 @@ Datum unite_garray(PG_FUNCTION_ARGS)
 	for (i=0; i<nelems; i++)
 	{
 		PG_LWGEOM *geom = (PG_LWGEOM *)(ARR_DATA_PTR(array)+offset);
-		offset += INTALIGN(geom->size);
+		offset += INTALIGN(VARSIZE(geom));
 
 		pgis_geom = geom;
 
@@ -1339,7 +1341,10 @@ Datum overlaps(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(result);
 }
 
-int point_in_polygon(LWPOLY *polygon, LWPOINT *point);
+int point_in_polygon(RTREE_NODE **root, int ringCount, LWPOINT *point);
+int point_in_polygon_deprecated(LWPOLY *polygon, LWPOINT *point);
+int point_outside_polygon(RTREE_NODE **root, int ringCount, LWPOINT *point);
+int point_outside_polygon_deprecated(LWPOLY *polygon, LWPOINT *point);
 int point_in_multipolygon(LWMPOLY *mpolygon, LWPOINT *point);
 
 PG_FUNCTION_INFO_V1(contains);
@@ -1354,6 +1359,8 @@ Datum contains(PG_FUNCTION_ARGS)
         LWPOLY *poly;
         /* LWMPOLY *mpoly; */
         LWPOINT *point;
+        RTREE_POLY_CACHE *poly_cache;
+        MemoryContext old_context;
 
 #ifdef PROFILE
 	profstart(PROF_QRUN);
@@ -1389,12 +1396,24 @@ Datum contains(PG_FUNCTION_ARGS)
 #ifdef PGIS_DEBUG
                 lwnotice("Point in Polygon test requested...short-circuiting.");
 #endif
+
                 poly = lwpoly_deserialize(SERIALIZED_FORM(geom1));
                 point = lwpoint_deserialize(SERIALIZED_FORM(geom2));
 #ifdef PGIS_DEBUG
                 lwnotice("Precall point_in_polygon %p, %p", poly, point);
 #endif
-                if(point_in_polygon(poly, point) == 0)
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom1), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+                if(point_in_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
                 {
 	                PG_FREE_IF_COPY(geom1, 0);
 	                PG_FREE_IF_COPY(geom2, 1);
@@ -1489,6 +1508,145 @@ Datum contains(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(result);
 }
 
+/*
+ * Described at 
+ * http://lin-ear-th-inking.blogspot.com/2007/06/subtleties-of-ogc-covers-spatial.html
+ */
+PG_FUNCTION_INFO_V1(covers);
+Datum covers(PG_FUNCTION_ARGS)
+{
+	PG_LWGEOM *geom1;
+	PG_LWGEOM *geom2;
+	GEOSGeom g1,g2;
+	bool result;
+	BOX2DFLOAT4 box1, box2;
+        int type1, type2;
+        LWPOLY *poly;
+        /* LWMPOLY *mpoly; */
+        LWPOINT *point;
+        RTREE_POLY_CACHE *poly_cache;
+        MemoryContext old_context;
+        char *patt = "******FF*";
+
+#ifdef PROFILE
+	profstart(PROF_QRUN);
+#endif
+
+	geom1 = (PG_LWGEOM *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
+	geom2 = (PG_LWGEOM *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
+
+	errorIfGeometryCollection(geom1,geom2);
+	errorIfSRIDMismatch(pglwgeom_getSRID(geom1), pglwgeom_getSRID(geom2));
+
+	/*
+	 * short-circuit 1: if geom2 bounding box is not completely inside
+	 * geom1 bounding box we can prematurely return FALSE.
+	 * Do the test IFF BOUNDING BOX AVAILABLE.
+	 */
+	if ( getbox2d_p(SERIALIZED_FORM(geom1), &box1) &&
+		getbox2d_p(SERIALIZED_FORM(geom2), &box2) )
+	{
+		if ( box2.xmin < box1.xmin ) PG_RETURN_BOOL(FALSE);
+		if ( box2.xmax > box1.xmax ) PG_RETURN_BOOL(FALSE);
+		if ( box2.ymin < box1.ymin ) PG_RETURN_BOOL(FALSE);
+		if ( box2.ymax > box1.ymax ) PG_RETURN_BOOL(FALSE);
+	}
+        /*
+         * short-circuit 2: if geom2 is a point and geom1 is a polygon
+         * call the point-in-polygon function.
+         */
+        type1 = lwgeom_getType((uchar)SERIALIZED_FORM(geom1)[0]);
+        type2 = lwgeom_getType((uchar)SERIALIZED_FORM(geom2)[0]);
+        if(type1 == POLYGONTYPE && type2 == POINTTYPE)
+        {
+#ifdef PGIS_DEBUG
+                lwnotice("Point in Polygon test requested...short-circuiting.");
+#endif
+
+                poly = lwpoly_deserialize(SERIALIZED_FORM(geom1));
+                point = lwpoint_deserialize(SERIALIZED_FORM(geom2));
+#ifdef PGIS_DEBUG
+                lwnotice("Precall point_in_polygon %p, %p", poly, point);
+#endif
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom1), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+                if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
+                {
+	                PG_FREE_IF_COPY(geom1, 0);
+	                PG_FREE_IF_COPY(geom2, 1);
+                        lwgeom_release((LWGEOM *)poly);
+                        lwgeom_release((LWGEOM *)point);
+                        PG_RETURN_BOOL(TRUE);
+                }
+                else
+                {
+                        PG_FREE_IF_COPY(geom1, 0);
+                        PG_FREE_IF_COPY(geom2, 1);
+                        lwgeom_release((LWGEOM *)poly);
+                        lwgeom_release((LWGEOM *)point);
+                        PG_RETURN_BOOL(FALSE);
+                }
+        } 
+        else 
+        {
+#ifdef PGIS_DEBUG
+                lwnotice("Covers: type1: %d, type2: %d", type1, type2);
+#endif
+        }
+        
+	initGEOS(lwnotice, lwnotice);
+
+#ifdef PROFILE
+	profstart(PROF_P2G1);
+#endif
+	g1 = POSTGIS2GEOS(geom1);
+#ifdef PROFILE
+	profstop(PROF_P2G1);
+#endif
+#ifdef PROFILE
+	profstart(PROF_P2G2);
+#endif
+	g2 = POSTGIS2GEOS(geom2);
+#ifdef PROFILE
+	profstop(PROF_P2G2);
+#endif
+
+#ifdef PROFILE
+	profstart(PROF_GRUN);
+#endif
+	result = GEOSRelatePattern(g1,g2,patt);
+#ifdef PROFILE
+	profstop(PROF_GRUN);
+#endif
+
+	GEOSGeom_destroy(g1);
+	GEOSGeom_destroy(g2);
+
+	if (result == 2)
+	{
+		elog(ERROR,"GEOS covers() threw an error!");
+		PG_RETURN_NULL(); /* never get here */
+	}
+
+#ifdef PROFILE
+	profstop(PROF_QRUN);
+	profreport("geos",geom1, geom2, NULL);
+#endif
+
+	PG_FREE_IF_COPY(geom1, 0);
+	PG_FREE_IF_COPY(geom2, 1);
+
+	PG_RETURN_BOOL(result);
+}
 
 PG_FUNCTION_INFO_V1(within);
 Datum within(PG_FUNCTION_ARGS)
@@ -1501,6 +1659,8 @@ Datum within(PG_FUNCTION_ARGS)
         LWPOLY *poly;
         LWPOINT *point;
         int type1, type2;
+        MemoryContext old_context;
+        RTREE_POLY_CACHE *poly_cache;
 
 #ifdef PROFILE
 	profstart(PROF_QRUN);
@@ -1536,9 +1696,21 @@ Datum within(PG_FUNCTION_ARGS)
 #ifdef PGIS_DEBUG
                 lwnotice("Point in Polygon test requested...short-circuiting.");
 #endif
+
                 point = lwpoint_deserialize(SERIALIZED_FORM(geom1));
                 poly = lwpoly_deserialize(SERIALIZED_FORM(geom2));
-                if(point_in_polygon(poly, point) == 0)
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom2), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+                if(point_in_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
                 {
 	                PG_FREE_IF_COPY(geom1, 0);
 	                PG_FREE_IF_COPY(geom2, 1);
@@ -1602,6 +1774,141 @@ Datum within(PG_FUNCTION_ARGS)
 }
 
 
+/*
+ * Described at:
+ * http://lin-ear-th-inking.blogspot.com/2007/06/subtleties-of-ogc-covers-spatial.html
+ */
+PG_FUNCTION_INFO_V1(coveredby);
+Datum coveredby(PG_FUNCTION_ARGS)
+{
+	PG_LWGEOM *geom1;
+	PG_LWGEOM *geom2;
+	GEOSGeom g1,g2;
+	bool result;
+	BOX2DFLOAT4 box1, box2;
+        LWPOLY *poly;
+        LWPOINT *point;
+        int type1, type2;
+        MemoryContext old_context;
+        RTREE_POLY_CACHE *poly_cache;
+        char *patt = "**F**F***";
+
+#ifdef PROFILE
+	profstart(PROF_QRUN);
+#endif
+
+	geom1 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
+	geom2 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
+
+	errorIfGeometryCollection(geom1,geom2);
+	errorIfSRIDMismatch(pglwgeom_getSRID(geom1), pglwgeom_getSRID(geom2));
+
+	/*
+	 * short-circuit 1: if geom1 bounding box is not completely inside
+	 * geom2 bounding box we can prematurely return FALSE.
+	 * Do the test IFF BOUNDING BOX AVAILABLE.
+	 */
+	if ( getbox2d_p(SERIALIZED_FORM(geom1), &box1) &&
+		getbox2d_p(SERIALIZED_FORM(geom2), &box2) )
+	{
+		if ( box1.xmin < box2.xmin ) PG_RETURN_BOOL(FALSE);
+		if ( box1.xmax > box2.xmax ) PG_RETURN_BOOL(FALSE);
+		if ( box1.ymin < box2.ymin ) PG_RETURN_BOOL(FALSE);
+		if ( box1.ymax > box2.ymax ) PG_RETURN_BOOL(FALSE);
+
+#ifdef PGIS_DEBUG
+                lwnotice("bounding box short-circuit missed.");
+#endif
+	}
+        /*
+         * short-circuit 2: if geom1 is a point and geom2 is a polygon
+         * call the point-in-polygon function.
+         */
+        type1 = lwgeom_getType((uchar)SERIALIZED_FORM(geom1)[0]);
+        type2 = lwgeom_getType((uchar)SERIALIZED_FORM(geom2)[0]);
+        if(type1 == POINTTYPE && type2 == POLYGONTYPE)
+        {
+#ifdef PGIS_DEBUG
+                lwnotice("Point in Polygon test requested...short-circuiting.");
+#endif
+
+                point = lwpoint_deserialize(SERIALIZED_FORM(geom1));
+                poly = lwpoly_deserialize(SERIALIZED_FORM(geom2));
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom2), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+                if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
+                {
+	                PG_FREE_IF_COPY(geom1, 0);
+	                PG_FREE_IF_COPY(geom2, 1);
+                        lwgeom_release((LWGEOM *)poly);
+                        lwgeom_release((LWGEOM *)point);
+                        PG_RETURN_BOOL(TRUE);
+                }
+                else
+                {
+                        PG_FREE_IF_COPY(geom1, 0);
+                        PG_FREE_IF_COPY(geom2, 1);
+                        lwgeom_release((LWGEOM *)poly);
+                        lwgeom_release((LWGEOM *)point);
+                        PG_RETURN_BOOL(FALSE);
+                }
+        }
+
+	initGEOS(lwnotice, lwnotice);
+
+#ifdef PROFILE
+	profstart(PROF_P2G1);
+#endif
+	g1 = POSTGIS2GEOS(geom1);
+#ifdef PROFILE
+	profstop(PROF_P2G1);
+#endif
+#ifdef PROFILE
+	profstart(PROF_P2G2);
+#endif
+	g2 = POSTGIS2GEOS(geom2);
+#ifdef PROFILE
+	profstop(PROF_P2G2);
+#endif
+
+#ifdef PROFILE
+	profstart(PROF_GRUN);
+#endif
+	result = GEOSRelatePattern(g1,g2,patt);
+#ifdef PROFILE
+	profstop(PROF_GRUN);
+#endif
+
+	GEOSGeom_destroy(g1);
+	GEOSGeom_destroy(g2);
+
+	if (result == 2)
+	{
+		elog(ERROR,"GEOS coveredby() threw an error!");
+		PG_RETURN_NULL(); /* never get here */
+	}
+
+#ifdef PROFILE
+	profstop(PROF_QRUN);
+	profreport("geos",geom1, geom2, NULL);
+#endif
+
+	PG_FREE_IF_COPY(geom1, 0);
+	PG_FREE_IF_COPY(geom2, 1);
+
+	PG_RETURN_BOOL(result);
+}
+
+
 
 PG_FUNCTION_INFO_V1(crosses);
 Datum crosses(PG_FUNCTION_ARGS)
@@ -1691,6 +1998,11 @@ Datum intersects(PG_FUNCTION_ARGS)
 	GEOSGeom g1,g2;
 	bool result;
 	BOX2DFLOAT4 box1, box2;
+	int type1, type2;
+	LWPOINT *point;
+	LWPOLY *poly;
+        MemoryContext old_context;
+        RTREE_POLY_CACHE *poly_cache;
 
 #ifdef PROFILE
 	profstart(PROF_QRUN);
@@ -1716,6 +2028,85 @@ Datum intersects(PG_FUNCTION_ARGS)
 		if ( box2.ymin > box2.ymax ) PG_RETURN_BOOL(FALSE);
 	}
 
+	/*
+	 * short-circuit 2: if the geoms are a point and a polygon,
+	 * call the point_outside_polygon function.
+	 */
+	type1 = lwgeom_getType((uchar)SERIALIZED_FORM(geom1)[0]);
+	type2 = lwgeom_getType((uchar)SERIALIZED_FORM(geom2)[0]);
+	if(type1 == POINTTYPE && type2 == POLYGONTYPE)
+	{
+#ifdef PGIS_DEBUG
+		lwnotice("Point in Polygon test requested...short-circuiting.");
+#endif
+		point = lwpoint_deserialize(SERIALIZED_FORM(geom1));
+		poly = lwpoly_deserialize(SERIALIZED_FORM(geom2));
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom2), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+		if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 1);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(TRUE);
+		}
+		else
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 1);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(FALSE);
+		}
+	}
+	else if(type1 == POLYGONTYPE && type2 == POINTTYPE)
+	{
+	#ifdef PGIS_DEBUG
+		lwnotice("Point in Polygon test requested...short-circuiting.");
+#endif
+		point = lwpoint_deserialize(SERIALIZED_FORM(geom2));
+		poly = lwpoly_deserialize(SERIALIZED_FORM(geom1));
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom1), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+		if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0)
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 1);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(TRUE);
+		}
+		else
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 1);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(FALSE);
+		}
+
+	}
+
+
 	initGEOS(lwnotice, lwnotice);
 
 #ifdef PROFILE
@@ -1847,6 +2238,11 @@ Datum disjoint(PG_FUNCTION_ARGS)
 	GEOSGeom g1,g2;
 	bool result;
 	BOX2DFLOAT4 box1, box2;
+	int type1, type2;
+	LWPOLY *poly;
+	LWPOINT *point;
+        MemoryContext old_context;
+        RTREE_POLY_CACHE *poly_cache;
 
 #ifdef PROFILE
 	profstart(PROF_QRUN);
@@ -1872,6 +2268,83 @@ Datum disjoint(PG_FUNCTION_ARGS)
 		if ( box2.ymin > box1.ymax ) PG_RETURN_BOOL(TRUE);
 	}
 
+	/*
+	 * short-circuit 2: if the geoms are a point and a polygon, 
+	 * call the point_outside_polygon function.
+	 */
+	type1 = lwgeom_getType((uchar)SERIALIZED_FORM(geom1)[0]);
+	type2 = lwgeom_getType((uchar)SERIALIZED_FORM(geom2)[0]);
+	if(type1 == POINTTYPE && type2 == POLYGONTYPE)
+	{
+#ifdef PGIS_DEBUG
+		lwnotice("Point outside Polygon test requested...short-circuiting.");
+#endif
+		point = lwpoint_deserialize(SERIALIZED_FORM(geom1));
+		poly = lwpoly_deserialize(SERIALIZED_FORM(geom2));
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom2), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+		if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0) 
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 0);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(FALSE);
+		}
+		else
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 0);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(TRUE);
+		}
+	}
+	else if(type1 == POLYGONTYPE && type2 == POINTTYPE)
+	{
+#ifdef PGIS_DEBUG
+		lwnotice("Point outside Polygon test requested...short-circuiting.");
+#endif
+		point = lwpoint_deserialize(SERIALIZED_FORM(geom2));
+		poly = lwpoly_deserialize(SERIALIZED_FORM(geom1));
+
+                /*
+                 * Switch the context to the function-scope context,
+                 * retrieve the appropriate cache object, cache it for 
+                 * future use, then switch back to the local context.
+                 */                 
+                old_context = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+                poly_cache = retrieveCache(poly, SERIALIZED_FORM(geom2), fcinfo->flinfo->fn_extra);
+                fcinfo->flinfo->fn_extra = poly_cache;
+                MemoryContextSwitchTo(old_context);
+
+		if(point_outside_polygon(poly_cache->ringIndices, poly_cache->ringCount, point) == 0) 
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 0);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(FALSE);
+		}
+		else
+		{
+			PG_FREE_IF_COPY(geom1, 0);
+			PG_FREE_IF_COPY(geom2, 0);
+			lwgeom_release((LWGEOM *)poly);
+			lwgeom_release((LWGEOM *)point);
+			PG_RETURN_BOOL(TRUE);
+		}
+	}
+
 	initGEOS(lwnotice, lwnotice);
 
 #ifdef PROFILE
@@ -2070,7 +2543,7 @@ Datum relate_full(PG_FUNCTION_ARGS)
 	len = strlen(relate_str) + VARHDRSZ;
 
 	result= palloc(len);
-	VARATT_SIZEP(result) = len;
+	SET_VARSIZE(result, len);
 
 	memcpy(VARDATA(result), relate_str, len-VARHDRSZ);
 
@@ -2600,8 +3073,11 @@ LWGEOM2GEOS(LWGEOM *lwgeom)
 {
 	GEOSCoordSeq sq;
 	GEOSGeom g, shell, *geoms;
+	/*
+        LWGEOM *tmp;
+	*/
 	unsigned int ngeoms, i;
-	int type = TYPE_GETTYPE(lwgeom->type);
+	int type;
 	int geostype;
 #ifdef PGIS_DEBUG_POSTGIS2GEOS 
 	char *wkt;
@@ -2611,6 +3087,21 @@ LWGEOM2GEOS(LWGEOM *lwgeom)
 	lwnotice("LWGEOM2GEOS got a %s", lwgeom_typename(type));
 #endif
 
+        if(has_arc(lwgeom))
+        {
+#ifdef PGIS_DEBUG_CALLS
+                lwnotice("LWGEOM2GEOS_c: arced geometry found.");
+#endif
+		lwerror("Exception in LWGEOM2GEOS: curved geometry not supported.");
+		/*
+                tmp = lwgeom;
+                lwgeom = lwgeom_segmentize(tmp, 32);
+#ifdef PGIS_DEBUG_CALLS
+                lwnotice("LWGEOM2GEOM_c: was %p, is %p", tmp, lwgeom);
+#endif
+		*/
+        }
+        type = TYPE_GETTYPE(lwgeom->type);
 	switch (type)
 	{
 		LWPOINT *lwp;
@@ -2678,7 +3169,11 @@ LWGEOM2GEOS(LWGEOM *lwgeom)
 			break;
 
 		default:
+#ifdef PGIS_DEBUG
+                        lwerror("LWGEOM2GEOS_c: Unknown geometry type: %d", type);
+#else 
 			lwerror("Unknown geometry type: %d", type);
+#endif
 			return NULL;
 	}
 
@@ -2687,6 +3182,9 @@ LWGEOM2GEOS(LWGEOM *lwgeom)
 #ifdef PGIS_DEBUG_POSTGIS2GEOS 
 	wkt = GEOSGeomToWKT(g);
 	lwnotice("LWGEOM2GEOS: GEOSGeom: %s", wkt);
+	/*
+        if(tmp != NULL) lwgeom_release(tmp);
+	*/
 	free(wkt);
 #endif
 
@@ -2791,7 +3289,7 @@ Datum polygonize_garray(PG_FUNCTION_ARGS)
 	for (i=0; i<nelems; i++)
 	{
 		PG_LWGEOM *geom = (PG_LWGEOM *)(ARR_DATA_PTR(array)+offset);
-		offset += INTALIGN(geom->size);
+		offset += INTALIGN(VARSIZE(geom));
 
 		vgeoms[i] = POSTGIS2GEOS(geom);
 		if ( ! i )
