@@ -13,10 +13,12 @@ Datum relate_full(PG_FUNCTION_ARGS);
 Datum relate_pattern(PG_FUNCTION_ARGS);
 Datum disjoint(PG_FUNCTION_ARGS);
 Datum touches(PG_FUNCTION_ARGS);
+Datum intersects_reprecated(PG_FUNCTION_ARGS);
 Datum intersects(PG_FUNCTION_ARGS);
 Datum crosses(PG_FUNCTION_ARGS);
 Datum within(PG_FUNCTION_ARGS);
 Datum contains(PG_FUNCTION_ARGS);
+Datum covers(PG_FUNCTION_ARGS);
 Datum overlaps(PG_FUNCTION_ARGS);
 Datum isvalid(PG_FUNCTION_ARGS);
 Datum buffer(PG_FUNCTION_ARGS);
@@ -46,9 +48,10 @@ Datum linemerge(PG_FUNCTION_ARGS);
 //#define PGIS_DEBUG_CONVERTER 1
 #ifdef PGIS_DEBUG_CONVERTER
 #define PGIS_DEBUG_POSTGIS2JTS 1
-#define PGIS_DEBUG_JTS2POSTGIS 1
+#define PGIS_DEBUG_JTS2POSTGIS 1 
 #endif // PGIS_DEBUG_CONVERTER
 //#define PGIS_DEBUG 1
+//#define PGIS_DEBUG_POSTGIS2JTS
 
 /*
  * If you're having problems with JTS<->POSTGIS conversions
@@ -58,7 +61,8 @@ Datum linemerge(PG_FUNCTION_ARGS);
 #define WKT_P2J 0
 
 typedef void (*noticefunc)(const char *fmt, ...);
-typedef  struct JTSGeometry JTSGeometry;
+typedef struct JTSGeometry JTSGeometry;
+typedef struct JTSPrepGeometry JTSPrepGeometry;
 
 extern const char * createJTSPoint(POINT3D *pt);
 extern void initJTS(noticefunc);
@@ -126,7 +130,20 @@ LWLINE *lwline_from_geometry(JTSGeometry *g, char want3d);
 LWPOLY *lwpoly_from_geometry(JTSGeometry *g, char want3d);
 LWCOLLECTION *lwcollection_from_geometry(JTSGeometry *geom, char want3d);
 LWGEOM *JTS2LWGEOM(JTSGeometry *g, char want3d);
+void JTSPrintPreparedGeometry(JTSGeometry *pg);
+JTSPrepGeometry *JTSGetRHSCache();
+JTSPrepGeometry *JTSGetLHSCache();
+void JTSPutRHSCache(JTSPrepGeometry *rhs);
+void JTSPutLHSCache(JTSPrepGeometry *lhs);
 
+typedef struct
+{
+        uchar *serializedGeomA;
+        uchar *serializedGeomB;
+} PREPARED_GEOMETRY_CACHE;
+
+char JTSpreparedIntersects(JTSPrepGeometry *pg, JTSGeometry *g);
+JTSPrepGeometry *JTSPrepareGeometry(JTSGeometry *g);
 
 /*
  * Using BUFFER(0) version of unite_garray takes
@@ -189,7 +206,7 @@ Datum unite_garray(PG_FUNCTION_ARGS)
 	for (i=0; i<nelems; i++)
 	{
 		PG_LWGEOM *geom = (PG_LWGEOM *)(ARR_DATA_PTR(array)+offset);
-		offset += INTALIGN(geom->size);
+		offset += INTALIGN(VARSIZE(geom));
 
 		//lwnotice("unite_garray starting iteration %d of %d", i, nelems);
 
@@ -337,7 +354,7 @@ Datum unite_garray(PG_FUNCTION_ARGS)
 	for (i=0; i<nelems; i++)
 	{
 		PG_LWGEOM *geom = (PG_LWGEOM *)(ARR_DATA_PTR(array)+offset);
-		offset += INTALIGN(geom->size);
+		offset += INTALIGN(VARSIZE(geom));
 
 		pgis_geom = geom;
 
@@ -1388,6 +1405,87 @@ Datum contains(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(result);
 }
 
+/*
+ * Described at 
+ * http://lin-ear-th-inking.blogspot.com/2007/06/subtleties-of-ogc-covers-spatial.html
+ */
+PG_FUNCTION_INFO_V1(covers);
+Datum covers(PG_FUNCTION_ARGS)
+{
+	PG_LWGEOM *geom1;
+	PG_LWGEOM *geom2;
+	JTSGeometry *g1,*g2;
+	bool result;
+	BOX2DFLOAT4 box1, box2;
+        char *patt = "******FF*";
+
+#ifdef PROFILE
+	profstart(PROF_QRUN);
+#endif
+
+	geom1 = (PG_LWGEOM *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
+	geom2 = (PG_LWGEOM *)  PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
+
+	errorIfJTSGeometryCollection(geom1,geom2);
+	errorIfSRIDMismatch(pglwgeom_getSRID(geom1), pglwgeom_getSRID(geom2));
+
+	/*
+	 * short-circuit 1: if geom2 bounding box is not completely inside
+	 * geom1 bounding box we can prematurely return FALSE.
+	 * Do the test IFF BOUNDING BOX AVAILABLE.
+	 */
+	if ( getbox2d_p(SERIALIZED_FORM(geom1), &box1) &&
+		getbox2d_p(SERIALIZED_FORM(geom2), &box2) )
+	{
+		if ( box2.xmin < box1.xmin ) PG_RETURN_BOOL(FALSE);
+		if ( box2.xmax > box1.xmax ) PG_RETURN_BOOL(FALSE);
+		if ( box2.ymin < box1.ymin ) PG_RETURN_BOOL(FALSE);
+		if ( box2.ymax > box1.ymax ) PG_RETURN_BOOL(FALSE);
+	}
+
+	initJTS(lwnotice);
+
+#ifdef PROFILE
+	profstart(PROF_P2G1);
+#endif
+	g1 = POSTGIS2JTS(geom1);
+#ifdef PROFILE
+	profstop(PROF_P2G1);
+#endif
+#ifdef PROFILE
+	profstart(PROF_P2G2);
+#endif
+	g2 = POSTGIS2JTS(geom2);
+#ifdef PROFILE
+	profstop(PROF_P2G2);
+#endif
+
+#ifdef PROFILE
+	profstart(PROF_GRUN);
+#endif
+	result = JTSrelatePattern(g1,g2,patt);
+#ifdef PROFILE
+	profstop(PROF_GRUN);
+#endif
+
+        finishJTS();
+
+	if (result == 2)
+	{
+		elog(ERROR,"JTS covers() threw an error!");
+		PG_RETURN_NULL(); /* never get here */
+	}
+
+#ifdef PROFILE
+	profstop(PROF_QRUN);
+	profreport("geos",geom1, geom2, NULL);
+#endif
+
+	PG_FREE_IF_COPY(geom1, 0);
+	PG_FREE_IF_COPY(geom2, 1);
+
+	PG_RETURN_BOOL(result);
+}
 
 PG_FUNCTION_INFO_V1(within);
 Datum within(PG_FUNCTION_ARGS)
@@ -1466,7 +1564,90 @@ Datum within(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(result);
 }
 
+/*
+ * Described at:
+ * http://lin-ear-th-inking.blogspot.com/2007/06/subtleties-of-ogc-covers-spatial.html
+ */
+PG_FUNCTION_INFO_V1(coveredby);
+Datum coveredby(PG_FUNCTION_ARGS)
+{
+	PG_LWGEOM *geom1;
+	PG_LWGEOM *geom2;
+	JTSGeometry *g1,*g2;
+	bool result;
+	BOX2DFLOAT4 box1, box2;
+        char *patt = "**F**F***";
 
+#ifdef PROFILE
+	profstart(PROF_QRUN);
+#endif
+
+	geom1 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
+	geom2 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
+
+	errorIfJTSGeometryCollection(geom1,geom2);
+	errorIfSRIDMismatch(pglwgeom_getSRID(geom1), pglwgeom_getSRID(geom2));
+
+	/*
+	 * short-circuit 1: if geom1 bounding box is not completely inside
+	 * geom2 bounding box we can prematurely return FALSE.
+	 * Do the test IFF BOUNDING BOX AVAILABLE.
+	 */
+	if ( getbox2d_p(SERIALIZED_FORM(geom1), &box1) &&
+		getbox2d_p(SERIALIZED_FORM(geom2), &box2) )
+	{
+		if ( box1.xmin < box2.xmin ) PG_RETURN_BOOL(FALSE);
+		if ( box1.xmax > box2.xmax ) PG_RETURN_BOOL(FALSE);
+		if ( box1.ymin < box2.ymin ) PG_RETURN_BOOL(FALSE);
+		if ( box1.ymax > box2.ymax ) PG_RETURN_BOOL(FALSE);
+
+#ifdef PGIS_DEBUG
+                lwnotice("bounding box short-circuit missed.");
+#endif
+	}
+	initJTS(lwnotice);
+
+#ifdef PROFILE
+	profstart(PROF_P2G1);
+#endif
+	g1 = POSTGIS2JTS(geom1);
+#ifdef PROFILE
+	profstop(PROF_P2G1);
+#endif
+#ifdef PROFILE
+	profstart(PROF_P2G2);
+#endif
+	g2 = POSTGIS2JTS(geom2);
+#ifdef PROFILE
+	profstop(PROF_P2G2);
+#endif
+
+#ifdef PROFILE
+	profstart(PROF_GRUN);
+#endif
+	result = JTSrelatePattern(g1,g2,patt);
+#ifdef PROFILE
+	profstop(PROF_GRUN);
+#endif
+
+        finishJTS();
+
+	if (result == 2)
+	{
+		elog(ERROR,"JTS coveredby() threw an error!");
+		PG_RETURN_NULL(); /* never get here */
+	}
+
+#ifdef PROFILE
+	profstop(PROF_QRUN);
+	profreport("geos",geom1, geom2, NULL);
+#endif
+
+	PG_FREE_IF_COPY(geom1, 0);
+	PG_FREE_IF_COPY(geom2, 1);
+
+	PG_RETURN_BOOL(result);
+}
 
 PG_FUNCTION_INFO_V1(crosses);
 Datum crosses(PG_FUNCTION_ARGS)
@@ -1545,6 +1726,219 @@ Datum crosses(PG_FUNCTION_ARGS)
 	PG_RETURN_BOOL(result);
 }
 
+PG_FUNCTION_INFO_V1(intersects_reprecated);
+Datum intersects_reprecated(PG_FUNCTION_ARGS)
+{
+        PG_LWGEOM *geom1;
+        PG_LWGEOM *geom2;
+        bool result;
+        BOX2DFLOAT4 box1, box2;
+        MemoryContext oldContext;
+        PREPARED_GEOMETRY_CACHE *cache = NULL;
+        JTSPrepGeometry *preparedGeom = NULL;
+        JTSGeometry *jtsGeom = NULL;
+        int i, length;
+        uchar *serialized;
+
+#ifdef PGIS_DEBUG_CALLS
+        lwnotice("intersects (prepared jts geometry) called");
+#endif 
+
+        geom1 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(0));
+        geom2 = (PG_LWGEOM *)PG_DETOAST_DATUM(PG_GETARG_DATUM(1));
+
+        errorIfJTSGeometryCollection(geom1,geom2);
+        errorIfSRIDMismatch(pglwgeom_getSRID(geom1), pglwgeom_getSRID(geom2));
+
+        /*
+         * short-circuit 1: if geom2 bounding box does not overlap
+         * geom1 bounding box, we can prematurely return FALSE.
+         * Do the test IFF BOUNDING BOX AVAILABLE.
+         */
+        if(getbox2d_p(SERIALIZED_FORM(geom1), &box1) && 
+                getbox2d_p(SERIALIZED_FORM(geom2), &box2))
+        {
+                if(box2.xmax < box1.xmin) PG_RETURN_BOOL(FALSE);
+                if(box2.xmin > box1.xmax) PG_RETURN_BOOL(FALSE);
+                if(box2.ymax < box1.ymin) PG_RETURN_BOOL(FALSE);
+                if(box2.ymin > box1.ymax) PG_RETURN_BOOL(FALSE);
+        }
+
+	initJTS(lwnotice);
+
+        /*
+         * Any memory allocation that must survive beyond this function call
+         * must be made in the function context.  The terminology is a touch 
+         * confusing.  The function context will persist across all calls to 
+         * this function within the currently running query.
+         */
+        oldContext = MemoryContextSwitchTo(fcinfo->flinfo->fn_mcxt);
+        /*
+         * Find the cache.  If there is none, create it.
+         */
+
+#ifndef DISABLE_CACHE
+        cache = fcinfo->flinfo->fn_extra;
+#endif
+
+#ifdef PGIS_DEBUG
+        lwnotice("intersects: cache retrieved %p", cache);
+#endif
+        if(!cache)
+        {
+#ifdef PGIS_DEBUG
+                lwnotice("intersects: no cache found.  Creating one.");
+#endif
+                cache = lwalloc(sizeof(PREPARED_GEOMETRY_CACHE));
+                preparedGeom = JTSPrepareGeometry(POSTGIS2JTS(geom1));
+                
+                JTSPutLHSCache(preparedGeom);
+                length = lwgeom_size(SERIALIZED_FORM(geom1));
+                cache->serializedGeomA = lwalloc(length);
+                memcpy(cache->serializedGeomA, SERIALIZED_FORM(geom1), length);
+                jtsGeom = POSTGIS2JTS(geom2);
+                JTSPutRHSCache(JTSPrepareGeometry(jtsGeom));
+                length = lwgeom_size(SERIALIZED_FORM(geom2));
+                cache->serializedGeomB = lwalloc(length);
+                memcpy(cache->serializedGeomB, SERIALIZED_FORM(geom2), length);
+                fcinfo->flinfo->fn_extra = cache;
+#ifdef PGIS_DEBUG
+                lwnotice("Cache created containing geoms:");
+#endif
+        }
+        /*
+         * There is a cache.  We must determine which geometry, if either,
+         * is found in the cache.
+         */
+        else
+        {
+#ifdef PGIS_DEBUG
+                lwnotice("intersects: checking cache");
+#endif
+                serialized = SERIALIZED_FORM(geom1);
+                length = lwgeom_size(serialized);
+                if(lwgeom_size(cache->serializedGeomA) == length)
+                {
+                        for(i = 0; i < length; i++)
+                        {
+                                uchar a = serialized[i];
+                                uchar b = cache->serializedGeomA[i];
+                                if(a != b)
+                                        break;
+                        }
+                        if(i >= length)
+                        {
+#ifdef PGIS_DEBUG
+                                lwnotice("intersects: LHS geom found in cache.");
+#endif
+                                preparedGeom = JTSGetLHSCache();
+                                jtsGeom = POSTGIS2JTS(geom2);
+#ifdef PGIS_DEBUG
+                                lwnotice("Cache hit, geom:");
+                                lwnotice("preparedGeom %p, jtsGeom %p",
+                                        preparedGeom, jtsGeom);
+#endif
+                        }
+                }
+                
+                serialized = SERIALIZED_FORM(geom2);
+                length = lwgeom_size(serialized);
+                if(!preparedGeom && lwgeom_size(cache->serializedGeomB) == length)
+                {
+                        for(i = 0; i < length; i++)
+                        {
+                                uchar a = serialized[i];
+                                uchar b = cache->serializedGeomB[i];
+                                if(a != b)
+                                        break;
+                        }
+                        if(i >= length)
+                        {
+#ifdef PGIS_DEBUG
+                                lwnotice("intersects: RHS geom found in cache.");
+#endif
+                                preparedGeom = JTSGetRHSCache();
+                                jtsGeom = POSTGIS2JTS(geom1);
+#ifdef PGIS_DEBUG
+                                lwnotice("Cache hit, geom:");
+                                lwnotice("preparedGeom %p, jtsGeom %p",
+                                        preparedGeom, jtsGeom);
+#endif
+                        }
+                }
+
+                if(!preparedGeom)
+                {
+#ifdef PGIS_DEBUG
+                        lwnotice("intersects: cache miss.");
+#endif
+                        if(cache->serializedGeomA)
+                                lwfree(cache->serializedGeomA);
+                        preparedGeom = JTSPrepareGeometry(POSTGIS2JTS(geom1));
+                        JTSPutLHSCache(preparedGeom);
+                        length = lwgeom_size(SERIALIZED_FORM(geom1));
+                        cache->serializedGeomA = lwalloc(length);
+                        memcpy(cache->serializedGeomA, 
+                                SERIALIZED_FORM(geom1), length);
+
+                        if(cache->serializedGeomB)
+                                lwfree(cache->serializedGeomB);
+                        jtsGeom = POSTGIS2JTS(geom2);
+                        JTSPutRHSCache(JTSPrepareGeometry(jtsGeom));
+                        length = lwgeom_size(SERIALIZED_FORM(geom2));
+                        cache->serializedGeomB = lwalloc(length);
+                        memcpy(cache->serializedGeomB, 
+                                SERIALIZED_FORM(geom2), length);
+                }
+        }
+        MemoryContextSwitchTo(oldContext);
+#ifdef PGIS_DEBUG
+        lwnotice("intersects: prepared-%p, geom-%p", preparedGeom, jtsGeom);
+        JTSPrintPreparedGeometry(preparedGeom);
+#endif
+
+        if(!preparedGeom || !jtsGeom)
+        {
+
+                lwnotice("intersects error: Unable to prepare geometery (%p, %p)", preparedGeom, jtsGeom);
+	        result = JTSrelateIntersects(POSTGIS2JTS(geom1), 
+                        POSTGIS2JTS(geom2));
+                if (result == 2)
+                {
+                        lwerror("JTS intersects() threw an error!");
+                        PG_FREE_IF_COPY(geom1, 0);
+                        PG_FREE_IF_COPY(geom2, 1);
+                        PG_RETURN_NULL();
+                }
+                
+        }
+        else
+        {
+
+                result = JTSpreparedIntersects(preparedGeom, jtsGeom);
+
+                /*
+	        finishJTS();
+                */
+	        if (result == 2)
+	        {
+		        lwerror("JTS intersects() threw an error!");
+	                PG_FREE_IF_COPY(geom1, 0);
+	                PG_FREE_IF_COPY(geom2, 1);
+		        PG_RETURN_NULL(); //never get here
+	        }
+        }
+
+	PG_FREE_IF_COPY(geom1, 0);
+	PG_FREE_IF_COPY(geom2, 1);
+
+#ifdef PGIS_DEBUG_CALLS
+        lwnotice("intersects() returning.");
+#endif
+
+	PG_RETURN_BOOL(result);
+}
+
 
 
 PG_FUNCTION_INFO_V1(intersects);
@@ -1552,7 +1946,8 @@ Datum intersects(PG_FUNCTION_ARGS)
 {
 	PG_LWGEOM *geom1;
 	PG_LWGEOM *geom2;
-	JTSGeometry *g1,*g2;
+        JTSGeometry *g1;
+        JTSGeometry *g2;
 	bool result;
 	BOX2DFLOAT4 box1, box2;
 
@@ -1929,7 +2324,7 @@ Datum relate_full(PG_FUNCTION_ARGS)
 	len = strlen(relate_str) + VARHDRSZ;
 
 	result= palloc(len);
-	VARATT_SIZEP(result) = len;
+	SET_VARSIZE(result, len);
 	//*((int *) result) = len;
 
 	memcpy(VARDATA(result), relate_str, len-VARHDRSZ);
@@ -2326,6 +2721,7 @@ LWGEOM *
 JTS2LWGEOM(JTSGeometry *geom, char want3d)
 {
 	char *wkt;
+    SERIALIZED_LWGEOM *serialized_lwgeom;
 	PG_LWGEOM *pglwgeom;
 	LWGEOM *ret;
 
@@ -2337,7 +2733,7 @@ JTS2LWGEOM(JTSGeometry *geom, char want3d)
 	lwnotice("JTS2LWGEOM: wkt: %d bytes", strlen(wkt));
 #endif
 
-	pglwgeom = (PG_LWGEOM *)parse_lwgeom_wkt(wkt);
+    serialized_lwgeom = parse_lwgeom_wkt(wkt);
 
 	//lwnotice("JTS2LWGEOM(WKT): wkt parsed (got pglwgeom)");
 
@@ -2345,12 +2741,12 @@ JTS2LWGEOM(JTSGeometry *geom, char want3d)
 	lwnotice("JTS2LWGEOM: parsed: %s",
 		lwgeom_typename(TYPE_GETTYPE(pglwgeom->type)));
 #endif
-	ret = lwgeom_deserialize(SERIALIZED_FORM(pglwgeom));
+	ret = lwgeom_deserialize(serialized_lwgeom->lwgeom);
+
 	ret->SRID = JTSGetSRID(geom);
 	//lwnotice("JTS2LWGEOM: return SRID: %d", ret->SRID);
 
 	return ret;
-
 }
 
 #else // ! WKT_J2P
@@ -2536,6 +2932,13 @@ LWGEOM2JTS(LWGEOM *lwgeom)
 				collected, col->ngeoms, col->SRID,
 				TYPE_NDIMS(col->type)>2);
 
+                case CURVETYPE:
+                case COMPOUNDTYPE:
+                case CURVEPOLYTYPE:
+                case MULTICURVETYPE:
+                case MULTISURFACETYPE:
+                        lwerror("Exception in LWGEOM2JTS: curved geometry not supported.");
+                        return NULL;
 		default:
 			lwerror("Unknown geometry type: %d",
 				TYPE_GETTYPE(lwgeom->type));
@@ -2653,7 +3056,7 @@ Datum polygonize_garray(PG_FUNCTION_ARGS)
 	for (i=0; i<nelems; i++)
 	{
 		PG_LWGEOM *geom = (PG_LWGEOM *)(ARR_DATA_PTR(array)+offset);
-		offset += INTALIGN(geom->size);
+		offset += INTALIGN(VARSIZE(geom));
 
 		vgeoms[i] = POSTGIS2JTS(geom);
 		if ( ! i )
@@ -2791,7 +3194,7 @@ Datum postgis_jts_version(PG_FUNCTION_ARGS)
 
 	ver = JTSversion();
 	result = (text *) palloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	free(ver);
 
