@@ -1,5 +1,5 @@
 /**********************************************************************
- * $Id: lwgeom_transform.c 2553 2006-12-15 11:00:50Z strk $
+ * $Id: lwgeom_transform.c 2718 2007-11-27 22:19:30Z mcayland $
  *
  * PostGIS - Spatial Types for PostgreSQL
  * http://postgis.refractions.net
@@ -18,10 +18,12 @@
 
 #include "postgres.h"
 #include "fmgr.h"
+#include "miscadmin.h"
 
 #include "liblwgeom.h"
 #include "lwgeom_pg.h"
 
+
 Datum transform(PG_FUNCTION_ARGS);
 Datum transform_geom(PG_FUNCTION_ARGS);
 Datum postgis_proj_version(PG_FUNCTION_ARGS);
@@ -141,9 +143,16 @@ PJ *GetProjectionFromPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid);
 void AddToPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid, int other_srid);
 void DeleteFromPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid);
 
+/* Search path for PROJ.4 library */
+static bool IsPROJ4LibPathSet = false;
+void SetPROJ4LibPath();
+
 /* Memory context cache functions */
 static void PROJ4SRSCacheInit(MemoryContext context);
 static void PROJ4SRSCacheDelete(MemoryContext context);
+static void PROJ4SRSCacheReset(MemoryContext context);
+static bool PROJ4SRSCacheIsEmpty(MemoryContext context);
+static void PROJ4SRSCacheStats(MemoryContext context);
 #ifdef MEMORY_CONTEXT_CHECKING
 static void PROJ4SRSCacheCheck(MemoryContext context);
 #endif
@@ -156,11 +165,12 @@ static MemoryContextMethods PROJ4SRSCacheContextMethods = {
 	NULL,
  	NULL,
  	PROJ4SRSCacheInit,
- 	NULL,
- 	PROJ4SRSCacheDelete
+ 	PROJ4SRSCacheReset,
+ 	PROJ4SRSCacheDelete,
 #ifdef MEMORY_CONTEXT_CHECKING
- 	,PROJ4SRSCacheCheck
+ 	PROJ4SRSCacheCheck,
 #endif
+    PROJ4SRSCacheStats
 };
 #endif
 
@@ -170,10 +180,10 @@ static MemoryContextMethods PROJ4SRSCacheContextMethods = {
 	NULL,
 	NULL,
 	PROJ4SRSCacheInit,
-	NULL,
+	PROJ4SRSCacheReset,
 	PROJ4SRSCacheDelete,
 	NULL,
-	NULL
+	PROJ4SRSCacheStats
 #ifdef MEMORY_CONTEXT_CHECKING
  	,PROJ4SRSCacheCheck
 #endif
@@ -186,11 +196,11 @@ static MemoryContextMethods PROJ4SRSCacheContextMethods = {
 	NULL,
 	NULL,
 	PROJ4SRSCacheInit,
-	NULL,
+	PROJ4SRSCacheReset,
 	PROJ4SRSCacheDelete,
 	NULL,
-	NULL,
-	NULL
+	PROJ4SRSCacheIsEmpty,
+	PROJ4SRSCacheStats
 #ifdef MEMORY_CONTEXT_CHECKING
 	,PROJ4SRSCacheCheck
 #endif
@@ -229,6 +239,36 @@ PROJ4SRSCacheDelete(MemoryContext context)
 	DeletePJHashEntry(context);
 }
 
+static void
+PROJ4SRSCacheReset(MemoryContext context)
+{
+	/*
+	 * Do nothing, but we must supply a function since this call is mandatory according to tgl
+     * (see postgis-devel archives July 2007)
+	 */
+}
+
+static bool
+PROJ4SRSCacheIsEmpty(MemoryContext context)
+{
+	/*
+	 * Always return false since this call is mandatory according to tgl
+     * (see postgis-devel archives July 2007)
+	 */
+    return FALSE;
+}
+
+static void
+PROJ4SRSCacheStats(MemoryContext context)
+{
+	/*
+	 * Simple stats display function - we must supply a function since this call is mandatory according to tgl
+     * (see postgis-devel archives July 2007)
+	 */
+    
+    fprintf(stderr, "%s: PROJ4 context\n", context->name);
+}
+
 #ifdef MEMORY_CONTEXT_CHECKING
 static void
 PROJ4SRSCacheCheck(MemoryContext context)
@@ -277,7 +317,7 @@ static HTAB *CreatePJHash(void)
 	ctl.entrysize = sizeof(PJHashEntry);
 	ctl.hash = mcxt_ptr_hash;
 
-	return hash_create("PROJ4 Backend PJ MemoryContext Hash", PROJ4_BACKEND_HASH_SIZE, &ctl, (HASH_ELEM | HASH_FUNCTION));
+	return hash_create("PostGIS PROJ4 Backend PJ MemoryContext Hash", PROJ4_BACKEND_HASH_SIZE, &ctl, (HASH_ELEM | HASH_FUNCTION));
 }
 
 static void AddPJHashEntry(MemoryContext mcxt, PJ *projection)
@@ -393,6 +433,7 @@ AddToPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid, int other_srid)
 	PJ *projection = NULL;
 	char *proj_str;
 	char proj4_spi_buffer[256];
+	int* pj_errno_ref;
 
 	/* Connect */
 	spi_result = SPI_connect();
@@ -418,11 +459,12 @@ AddToPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid, int other_srid)
 		strcpy(proj_str, SPI_getvalue(tuple, tupdesc, 1));
 		projection = make_project(proj_str);
 
-		if ( (projection == NULL) || pj_errno)
+		pj_errno_ref = pj_get_errno_ref();
+		if ( (projection == NULL) || (*pj_errno_ref))
 		{
 			/* we need this for error reporting */
 			/*pfree(projection); */
-			elog(ERROR, "AddToPROJ4SRSCache: couldn't parse proj4 string: '%s': %s", proj_str, pj_strerrno(pj_errno));
+			elog(ERROR, "AddToPROJ4SRSCache: couldn't parse proj4 string: '%s': %s", proj_str, pj_strerrno(*pj_errno_ref));
 		}
 
 		/*
@@ -460,7 +502,7 @@ AddToPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid, int other_srid)
 		PJMemoryContext = MemoryContextCreate(T_AllocSetContext, 8192,
 					&PROJ4SRSCacheContextMethods,
 					PROJ4Cache->PROJ4SRSCacheContext,
-					"PROJ4 PJ Memory Context");
+					"PostGIS PROJ4 PJ Memory Context");
 
 		/* Create the backend hash if it doesn't already exist */
 		if (!PJHash)
@@ -526,6 +568,38 @@ void DeleteFromPROJ4SRSCache(PROJ4PortalCache *PROJ4Cache, int srid)
 }
 
 
+/*
+ * Specify an alternate directory for the PROJ.4 grid files
+ * (this should augment the PROJ.4 compile-time path)
+ *
+ * It's main purpose is to allow Win32 PROJ.4 installations
+ * to find a set grid shift files, although other platforms
+ * may find this useful too.
+ */
+void SetPROJ4LibPath()
+{
+	char *path;
+	const char **proj_lib_path;
+
+	/* 
+	 * Get the sharepath and append /contrib/postgis/nad to form a suitable
+	 * directory in which to store the grid shift files
+	 */	
+	proj_lib_path = palloc(sizeof(char *));
+	path = palloc(MAXPGPATH);
+	*proj_lib_path = path;
+
+	get_share_path(my_exec_path, path);
+	strncat(path, "/contrib/postgis/nad", MAXPGPATH - strlen(path) - 1);
+
+	/* Set the search path for PROJ.4 */
+	pj_set_searchpath(1, proj_lib_path);
+
+	/* Ensure we only do this once... */
+	IsPROJ4LibPathSet = true;
+}
+
+
 /*
  * This is *exactly* the same as PROJ.4's pj_transform(),
  * but it doesn't do the datum shift.
@@ -537,7 +611,7 @@ int point_offset, double *x, double *y, double *z )
     long      i;
     /*int       need_datum_shift; */
 
-    pj_errno = 0;
+    int* pj_errno_ref;
 
     if( point_offset == 0 )
         point_offset = 1;
@@ -553,8 +627,10 @@ int point_offset, double *x, double *y, double *z )
             projected_loc.v = y[point_offset*i];
 
             geodetic_loc = pj_inv( projected_loc, srcdefn );
-            if( pj_errno != 0 )
-                return pj_errno;
+            
+            pj_errno_ref = pj_get_errno_ref();
+            if( (*pj_errno_ref) != 0 )
+                return *pj_errno_ref;
 
             x[point_offset*i] = geodetic_loc.u;
             y[point_offset*i] = geodetic_loc.v;
@@ -572,8 +648,10 @@ int point_offset, double *x, double *y, double *z )
             geodetic_loc.v = y[point_offset*i];
 
             projected_loc = pj_fwd( geodetic_loc, dstdefn );
-            if( pj_errno != 0 )
-                return pj_errno;
+
+            pj_errno_ref = pj_get_errno_ref();
+            if( (*pj_errno_ref) != 0 )
+                return *pj_errno_ref;
 
             x[point_offset*i] = projected_loc.u;
             y[point_offset*i] = projected_loc.v;
@@ -731,6 +809,8 @@ lwgeom_transform_recursive(uchar *geom, PJ *inpj, PJ *outpj)
 }
 
 
+
+
 /*
  * transform( GEOMETRY, INT (output srid) )
  * tmpPts - if there is a nadgrid error (-38), we re-try the transform
@@ -765,6 +845,10 @@ Datum transform(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 	}
 
+	/* Set the search path if we haven't already */
+	if (!IsPROJ4LibPathSet)
+		SetPROJ4LibPath();
+
 	/*
 	 * If input SRID and output SRID are equal, return geometry
 	 * without transform it
@@ -876,6 +960,8 @@ Datum transform_geom(PG_FUNCTION_ARGS)
 	text *output_proj4_text;
 	int32 result_srid ;
 	uchar *srl;
+	int* pj_errno_ref;
+
 
 	result_srid = PG_GETARG_INT32(3);
 	if (result_srid == -1)
@@ -892,41 +978,49 @@ Datum transform_geom(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 	}
 
+	/* Set the search path if we haven't already */
+	if (!IsPROJ4LibPathSet)
+		SetPROJ4LibPath();
+
 	input_proj4_text  = (PG_GETARG_TEXT_P(1));
 	output_proj4_text = (PG_GETARG_TEXT_P(2));
 
-	input_proj4 = (char *)palloc(input_proj4_text->vl_len+1-4);
-	memcpy(input_proj4, input_proj4_text->vl_dat,
-		input_proj4_text->vl_len-4);
-	input_proj4[input_proj4_text->vl_len-4] = 0; /* null terminate */
+	input_proj4 = (char *)palloc(VARSIZE(input_proj4_text)+1-4);
+	memcpy(input_proj4, VARDATA(input_proj4_text),
+		VARSIZE(input_proj4_text)-VARHDRSZ);
+	input_proj4[VARSIZE(input_proj4_text)-VARHDRSZ] = 0; /* null terminate */
 
-	output_proj4 = (char *) palloc(output_proj4_text->vl_len +1-4);
-	memcpy(output_proj4, output_proj4_text->vl_dat,
-		output_proj4_text->vl_len-4);
-	output_proj4[output_proj4_text->vl_len-4] = 0; /* null terminate */
+	output_proj4 = (char *) palloc(VARSIZE(output_proj4_text) +1-VARHDRSZ);
+	memcpy(output_proj4, VARDATA(output_proj4_text),
+		VARSIZE(output_proj4_text)-VARHDRSZ);
+	output_proj4[VARSIZE(output_proj4_text)-VARHDRSZ] = 0; /* null terminate */
 
 	/* make input and output projection objects */
 	input_pj = make_project(input_proj4);
-	if ( (input_pj == NULL) || pj_errno)
+	
+	pj_errno_ref = pj_get_errno_ref();
+        if ( (input_pj == NULL) || (*pj_errno_ref))
 	{
 		/* we need this for error reporting */
 		/* pfree(input_proj4); */
 
 		pfree(output_proj4);
 		pfree(geom);
-		elog(ERROR, "transform: couldn't parse proj4 input string: '%s': %s", input_proj4, pj_strerrno(pj_errno));
+		elog(ERROR, "transform: couldn't parse proj4 input string: '%s': %s", input_proj4, pj_strerrno(*pj_errno_ref));
 		PG_RETURN_NULL();
 	}
 	pfree(input_proj4);
 
 	output_pj = make_project(output_proj4);
-	if ((output_pj == NULL)|| pj_errno)
+	
+	pj_errno_ref = pj_get_errno_ref();
+	if ((output_pj == NULL)|| (*pj_errno_ref))
 	{
 		/* we need this for error reporting */
 		/* pfree(output_proj4); */
 		pj_free(input_pj);
 		pfree(geom);
-		elog(ERROR, "transform: couldn't parse proj4 output string: '%s': %s", output_proj4, pj_strerrno(pj_errno));
+		elog(ERROR, "transform: couldn't parse proj4 output string: '%s': %s", output_proj4, pj_strerrno(*pj_errno_ref));
 		PG_RETURN_NULL();
 	}
 	pfree(output_proj4);
@@ -965,10 +1059,10 @@ Datum transform_geom(PG_FUNCTION_ARGS)
 PG_FUNCTION_INFO_V1(postgis_proj_version);
 Datum postgis_proj_version(PG_FUNCTION_ARGS)
 {
-	const char *ver = pj_release;
+	const char *ver = pj_get_release();
 	text *result;
 	result = (text *) palloc(VARHDRSZ  + strlen(ver));
-	VARATT_SIZEP(result) = VARHDRSZ + strlen(ver) ;
+	SET_VARSIZE(result, VARHDRSZ + strlen(ver));
 	memcpy(VARDATA(result), ver, strlen(ver));
 	PG_RETURN_POINTER(result);
 }
@@ -977,23 +1071,28 @@ Datum postgis_proj_version(PG_FUNCTION_ARGS)
 int
 transform_point(POINT4D *pt, PJ *srcpj, PJ *dstpj)
 {
+	int* pj_errno_ref;
+	
 	if (srcpj->is_latlong) to_rad(pt);
 	pj_transform(srcpj, dstpj, 1, 2, &(pt->x), &(pt->y), &(pt->z));
-	if (pj_errno)
+	
+	pj_errno_ref = pj_get_errno_ref();
+        if (*pj_errno_ref)
 	{
-		if (pj_errno == -38)  /*2nd chance */
+		if ((*pj_errno_ref) == -38)  /*2nd chance */
 		{
 			elog(WARNING, "transform: %i (%s)",
-				pj_errno, pj_strerrno(pj_errno));
+				*pj_errno_ref, pj_strerrno(*pj_errno_ref));
 			/*couldnt do nadshift - do it without the datum */
 			pj_transform_nodatum(srcpj, dstpj, 1, 2,
 				&(pt->x), &(pt->y), NULL);
 		}
 
-		if (pj_errno)
+		pj_errno_ref = pj_get_errno_ref();
+		if ((*pj_errno_ref))
 		{
 			elog(ERROR,"transform: couldn't project point: %i (%s)",
-				pj_errno, pj_strerrno(pj_errno));
+				*pj_errno_ref, pj_strerrno(*pj_errno_ref));
 			return 0;
 		}
 	}
